/**
  * All the MyNetLayer specific data structs and operations are in this file.
  */
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "MyNetLayer.h"
#include "MNLCfgFile.h"

/**
  * This is the current list of packets to be sent out.
  * The packets are stored in the order in which they are received from the higher layer.
  * This buffer wraps around.
  */
MNLBuf gSendBuf[MAX_BUFFERED_PKTS];

/**
  * The socket on which packets are transmitted to their destinations.
  */
int gTxSkt;

/**
  * The socket on which we listen for sendto()s from the application.
  */
int gListenSkt;


extern int gRTTVar;
extern int gRTTMean;

/**
  * The index of the slot that contains the next packet to be sent out.
  * Looks at the mToSendFlag of the MNLBuf at the index gSendBufNextToSend.
  * If it is 0, it means that there are no outstanding pkts to be sent. 
  * This is because we are always sending out in the order of packets contained
  * in gSendBuf.  For handling re-ordering, we swap the order of packets in
  * gSendBuf.  The sending function is thus not aware of re-ordering.
  */
int gSendBufNextToSend;

/**
  * The index of the next free slot in the send buffer.
  * If the MNLBuf.mToSendFlag at index gSendBufNextFreeSlot is 1, it means
  * that we are out of buffer space.  The higher layer is trying to 
  * send at a faster rate than we can support.
  */
int gSendBufNextFreeSlot;


void MNLBufInit(MNLBuf* ipMnlBuf) {
  ipMnlBuf->mToSendFlag = 0;
  ipMnlBuf->mPktLen = 0;
}

void SendBufInit() {
  int lBufIndex;
  for (lBufIndex = 0; lBufIndex < MAX_BUFFERED_PKTS; lBufIndex++) {
    MNLBufInit(&(gSendBuf[lBufIndex]));
  }
  gSendBufNextToSend = 0;
  gSendBufNextFreeSlot = 0;
}



/**
  * Initializes the UDP socket on which MNL listens for packets the application
  * wishes to send.
  * @param ipListenPort The port number on which MNL is listening is put into this parameter.
  * @return 0 if no error. 1 if some error occurred.
  */
int ListenSktInit() {
    
  int lYes = 1;
  struct sockaddr_in lMyAddr;
  socklen_t lMyAddrLen = sizeof(lMyAddr);

  
  gListenSkt = socket(PF_INET, SOCK_DGRAM, 0);
  if (gListenSkt == -1) {
    printf ("MNL: Unable to open UDP listen socket: %s\n",
            strerror(errno)
          );
    return 1;
  }

  if (setsockopt(
                  gListenSkt,
                  SOL_SOCKET, SO_REUSEADDR, 
                  &lYes, sizeof(int)
                ) == -1) {
    printf ("MNL: Unable to set SO_REUSEADDR option on udp listen socket: %s.\n",
              strerror (errno)
            );
    return 1;
  }
 
  // Bind to a random port
  lMyAddr.sin_family = AF_INET; 
  lMyAddr.sin_port = 0;
  lMyAddr.sin_addr.s_addr = INADDR_ANY;
  memset(&(lMyAddr.sin_zero), '\0', sizeof(lMyAddr.sin_zero));

  if (bind( gListenSkt , 
            (struct sockaddr *)&lMyAddr, 
            lMyAddrLen
            ) == -1) {

    printf ( "MNL: Unable to bind udp listen socket to a random port: %s\n",
              strerror (errno)
            );
    return 1;
  }

  // Get the port on which we are listening.
  if ( getsockname (
                    gListenSkt,
                    (struct sockaddr *) &lMyAddr, 
                    &lMyAddrLen
                    ) != -1) {

    printf("MNL: Listening on %s:%d\n",
                    inet_ntoa (lMyAddr.sin_addr),
                    ntohs (lMyAddr.sin_port)
                );

    return 0;
  } else {
    printf ("MNL: Error: Unable to get port number of listening socket :%s\n",
            strerror(errno)
          );
    return 1;
  }
 
  return 0;

}




/**
  * @return 0 if initialization was successful.  Else returns 1.
  */
int MNLInit(const char* ipCfgFile ) {
  
  gTxSkt = 0;
  gListenSkt = 0;

  // Pass a specific seed if you want deterministic random behaviour
  srand(time(NULL));


  if (ReadCfgFile(ipCfgFile) != 1) {
    printf ("MNL: MNLInit: Fatal Error: Unable to read config file %s\n", ipCfgFile);
    return 1;
  }
 
  SendBufInit();

  if (TxSktInit() != 0) {
    printf("MNL: MNLInit: Fatal Error: Unable to initialize sending socket\n");
    return 1;
  }

  if (ListenSktInit() != 0) {
    printf("MNL: MNLInit: Fatal Error: Unable to initialize listening socket\n");
    return 1;
  }

  MNLMainLoop();

 return 0;
}

void MNLDeInit() {
  printf ("MNL: DeInit\n");

  if (gTxSkt != 0) {
    close (gTxSkt);
  } else {
    gTxSkt = 0;
  }

  if (gListenSkt != 0) {
    close (gListenSkt);
  } else {
    gListenSkt = 0;
  }
} 


